#!/usr/bin/python
# -*- coding: ascii -*-
"""
Stream utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import queue
import threading as th
import typing as tp

from .threading import SDK_TP, SimpleThreadTask


# -->> Definitions <<------------------


# -->> API <<--------------------------
class StreamRedirect:
    """
    Stream redirect utility implementation.
    Allows to store and redirect a given stream.
    """
    def __init__(self, name: str, stream_in: tp.IO[tp.AnyStr], stream_out: tp.IO[tp.AnyStr]) -> None:
        """
        Class initialization.

        :param str name:                Stream redirection name (used for naming threads).
        :param IO[AnyStr] stream_in:    Input stream. Will be stored on buffer and redirected to output.
        :param IO[AnyStr] stream_out:   Output stream.
        """
        # Configure redirection operation
        self._src   = stream_in
        self._out   = stream_out
        self._queue = queue.Queue()
        self._buff  = []
        self._ready = th.Event()
        # Start
        SDK_TP.enqueue(SimpleThreadTask(name=f"{self.__class__.__name__}_{name}_read", task=self._read))
        SDK_TP.enqueue(SimpleThreadTask(name=f"{self.__class__.__name__}_{name}_write", task=self._write))

    @property
    def buffer(self) -> str:
        """
        Stream buffer.
        """
        return "".join(self._buff)

    def join(self) -> None:
        """
        Waits until the stream is finished to continue.
        """
        self._ready.wait()

    def _write(self) -> None:
        """
        Writes every line into the output stream and buffer.
        """
        for line in iter(self._queue.get, None):
            self._buff.append(line)
            self._out.write(line)
        self._ready.set()

    def _read(self) -> None:
        """
        Parses and copies every line of the input stream.
        """
        for line in iter(self._src.readline, b""):
            self._queue.put(line.decode())
        self._queue.put(None)
        self._src.close()



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Version handler class.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import abc
import datetime as dt
import os

import attr

from ..utils.common import TPPath
from ..utils.math import closest_multi
from ..utils.path import Path, FileTypeError
from ..utils.subprocess import execute
from ..utils.version import Version


# -->> Definitions <<------------------
#: Source path
PATH_THIS = Path(os.path.abspath(os.path.dirname(__file__)))
#: TeMPLate path
PATH_TMPL = PATH_THIS / "templates"


# -->> API <<--------------------------
class AbstractVersionExporter(abc.ABC):
    """
    Version exporter abstraction.
    This class implements the logic used to export the version to any language.
    """
    @abc.abstractmethod
    def export(self, version: Version, path: TPPath, author: str = "Unknown") -> None:
        """
        Exports the version to be used on a given language.

        :param Version version: Version to be exported.
        :param TPPath path:     Target file path.
        :param str author:      Exported file owner.
        """


class AbstractVersionUpdater(abc.ABC):
    """
    Version updater abstraction.
    This class implements the logic to update the version number.
    """
    @abc.abstractmethod
    def update(self, version: Version, path: TPPath) -> None:
        """
        Updates the version number.

        :param Version version: Version to be exported.
        :param TPPath path:     Target file path.
        """


class AbstractVersionStorage(abc.ABC):
    """
    Version storage abstraction.
    This class implements the logic to store/retrieve the base version number.
    """
    @abc.abstractmethod
    def load(self, version: Version, path: TPPath) -> None:
        """
        Loads the version info from the given file.

        :param Version version: Version that will be updated after load.
        :param TPPath path:     Source file path.
        """

    @abc.abstractmethod
    def save(self, version: Version, path: TPPath) -> None:
        """
        Saves the version info on the given file.

        :param Version version: Version to be stored.
        :param TPPath path:     Target file path.
        """


@attr.s
class VersionHandler(Version):
    """
    Version handler implementation.

    Yse this class to perform operations over the version number.
    """
    #: Handler used to export the version number.
    exporter:   AbstractVersionExporter = attr.ib(default=None)
    #: Handler used to save/load the version number.
    storage:    AbstractVersionStorage  = attr.ib(default=None)
    #: Handler used to update the version number.
    updater:    AbstractVersionUpdater  = attr.ib(default=None)

    def save(self, path: TPPath = os.getcwd()) -> None:
        """
        Stores the version number on the given path.

        :param TPPath path:     Target file path.
        """
        # Check handler existence
        if self.storage is None:
            raise RuntimeError("Version storage handler not provided!")
        # Execute
        self.storage.save(version=self, path=path)

    def load(self, path: TPPath = os.getcwd()) -> None:
        """
        Retrieves the version number from the given path.

        :param TPPath path:     Source file path.
        """
        # Check handler existence
        if self.storage is None:
            raise RuntimeError("Version storage handler not provided!")
        # Execute
        self.storage.load(version=self, path=path)

    def update(self, path: TPPath = os.getcwd()) -> None:
        """
        Updates the version number.

        :param TPPath path:     Source file path.
        """
        # Check handler existence
        if self.updater is None:
            raise RuntimeError("Version updater handler not provided!")
        # Execute
        self.updater.update(version=self, path=path)

    def export(self, path: TPPath = os.getcwd(), author: str = "Unknown") -> None:
        """
        Exports the version number.

        :param TPPath path:     Target file path.
        :param str author:      Target file owner.
        """
        # Check handler existence
        if self.exporter is None:
            raise RuntimeError("Version exporter handler not provided!")
        # Execute
        self.exporter.export(version=self, path=path, author=author)


class SimpleVersionStorage(AbstractVersionStorage):
    """
    Simple Version Storage.
    Stores and load the version number on a simple text file.
    """
    #: Default version storage filename.
    FILENAME = "version.txt"

    def load(self, version: Version, path: TPPath) -> None:
        """
        Loads the version number from a text file.

        :param Version version: Version to be loaded.
        :param TPPath path:     Path from which the version will be loaded.
        """
        # Check file
        path = Path.validate_file(path=path, none_ok=False, must_exist=True, default=self.FILENAME)

        # Load version from file
        with path.open(mode="r", encoding="utf-8") as file:
            version.parse(text=file.read())

    def save(self, version: Version, path: TPPath) -> None:
        """
        Saves the version number as text.

        :note: This operation re-writes the target file to only store the version number.

        :param Version version: Version to be stored.
        :param TPPath path:     Path in which the version will be stored.
        """
        # Check file
        path = Path.validate_file(path=path, none_ok=False, default=self.FILENAME)

        # Store version string
        with path.open(mode="w", encoding="utf-8") as file:
            file.write(f"{version}")


class GitBuildVersionUpdater(AbstractVersionUpdater):
    """
    Git Version Updater.
    Implements the logic to retrieve to update the version build number based on
    the latest Git commit.
    """
    #: No repo version build number.
    NO_BUILD = 999999

    def update(self, version: Version, path: TPPath) -> None:
        """
        Updates the version build number based on the latest Git commit.

        :param Version version: Version to be updated.
        :param TPPath path:     Path in which the repository is initialized.
        """
        # Update the build number with the commit number
        ret = execute(cmd="git rev-parse --short HEAD", cwd=f"{path}", pipe=False)
        ret = (ret.stderr + ret.stdout).lower().strip()
        version.build = self.NO_BUILD if ("not a git" in ret) else int(ret, 16)


class CCppVersionExporter(AbstractVersionExporter):
    """
    C/C++ Version Exporter.
    Implements the logic to export the version number to a C/C++ header file
    """
    #: Item print size.
    ITEMSIZE = 20
    #: Default exported version filename.
    FILENAME = "version.h"
    #: Supported header file extensions.
    SUFFIXES = [".h", ".hpp"]
    #: C/C++ header template.
    TEMPLATE = PATH_TMPL / "template_version_c.h"

    def export(self, version: Version, path: TPPath, author: str = "Unknown") -> None:
        """
        Exports the version number to a C/C++ header using a header template.

        :param Version version: Version to be exported.
        :param TPPath path:     Target file path.
        :param str author:      Exported file owner.
        """
        # Check file
        path = Path.validate_file(path=path, none_ok=False, default=self.FILENAME)
        if path.suffix.lower() not in self.SUFFIXES:
            raise FileTypeError(f"Header path doesnt have the right suffix ({self.SUFFIXES}): {path}.")

        # Generate header
        with path.open(mode="w", encoding="utf-8") as file:
            tmpl = self.TEMPLATE.open(mode="r", encoding="utf-8").read()
            file.write(tmpl.format(
                file=path.name, author=author,
                guard=f"_{path.stem}_H_".upper(),
                date=f"{dt.datetime.now().strftime('%m/%d/%Y %H:%M:%S')}",
                major=self._format(item=version.major),
                minor=self._format(item=version.minor),
                build=self._format(item=version.build),
                version=f"{version}"
                ))

    def _format(self, item: int) -> str:
        """
        Formats the version number item.

        :param int item: Version item value.

        :return: Item entry.
        :rtype: str
        """
        hexval = f"0x{item:0{closest_multi(ref=len(hex(item)[2:]), base=2)}X}U"
        return f"{hexval:{self.ITEMSIZE}s}   /* DEC: {str(item):<{self.ITEMSIZE}s} */"



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Version number implementation.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import re

import attr

from ..utils.common import TPAny, TPText

# -->> Tunables <<---------------------


# -->> Definitions <<------------------


# -->> API <<--------------------------
@attr.s
class Version:
    """
    Simple version definition.
    """
    #: Regex version pattern.
    REGEX_VER = re.compile(pattern=r"(([0-9]*\.){2,}([a-z0-9]{1,}))", flags=re.I)
    #: Regex HEX pattern.
    REGEX_HEX = re.compile(pattern=r"^((0x){0,1}([a-f0-9]{1,}))$", flags=re.I)
    #: Regex INT pattern.
    REGEX_INT = re.compile(pattern=r"^([0-9]{1,})$", flags=re.I)

    #: Version major
    major:      int = attr.ib(default=99, converter=int)
    #: Version minor
    minor:      int = attr.ib(default=0, converter=int)
    #: Version build
    build:      int = attr.ib(default=0, converter=int)
    #: Flag. If enabled the build is parsed/printed as HEX.
    hex_build:  bool = attr.ib(default=False, converter=bool)

    def __str__(self) -> str:
        """
        Version as string: major.minor.build
        """
        return f"{self.major}.{self.minor}.{hex(self.build)[2:] if self.hex_build else self.build}"

    def parse(self, text: TPAny) -> None:
        """
        Parses a version string.

        :param TPAny text:      Version string.

        :raises ValueError: Input is not a string or contents don't match a version pattern.
        """
        # Avoid not compatible types
        constraints = getattr(TPText, "__constraints__")
        if not isinstance(text, constraints):
            raise ValueError(f"Parameter with value '{text}' can't be converted to text.")

        # Ensure format and search
        text  = text if isinstance(text, str) else text.decode(errors="ignore")
        match = Version.REGEX_VER.search(string=text.strip())
        if match is None:
            raise ValueError(f"Unable to parse a valid version number from '{text}'.")

        # Parse: major and minor
        items = match.group().lower().split('.')
        self.major, self.minor = map(int, items[:-1])

        # Parse: build
        base  = 16 if self.hex_build else 10
        regex = self.REGEX_HEX if self.hex_build else self.REGEX_INT
        match = regex.search(string=items[-1])
        self.build = 0 if (match is None) else int(match.group(), base)

    @staticmethod
    def from_str(text: TPAny, hex_build: bool = False) -> 'Version':
        """
        Parses a version number from a string.

        :param TPAny text:      Version string.
        :param bool hex_build:  If true, assumes that the build section is in HEX format.

        :return: Version number.
        :rtype: Version

        :raises ValueError: Input is not a string or contents don't match a version pattern.
        """
        ver = Version(hex_build=hex_build)
        ver.parse(text=text)
        return ver



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Subprocess execution utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import subprocess as sp
import sys
import time

from .path import TPPath, Path
from .stream import StreamRedirect


# -->> Definitions <<------------------


# -->> API <<--------------------------
def execute(cmd: str, cwd: TPPath = None, log: TPPath = None, pipe: bool = True) -> sp.CompletedProcess:
    """
    Execute the given command as a subprocess.

    :param TPPath cmd:  Command to be executed.
    :param TPPath cwd:  Command working directory.
    :param TPPath log:  File to store the execution logs.
    :param bool pipe:   Enable pipe output to terminal.
    """
    # Check paths
    cwd = Path.validate_dir(path=cwd, none_ok=True)
    log = Path.validate_dir(path=log, none_ok=True)

    # Prepare
    with sp.Popen(cmd, cwd=cwd, shell=True, close_fds=True, stdout=sp.PIPE, stderr=sp.PIPE) as proc:
        # Execute
        if pipe:
            # Piping needed...
            # Print header
            print(f"Executing:\n{cmd}\nOutput:")
            # Set piping...
            s_err = StreamRedirect(name="stderr", stream_in=proc.stderr, stream_out=sys.stderr)
            s_out = StreamRedirect(name="stdout", stream_in=proc.stdout, stream_out=sys.stdout)
            # Wait for process and get data
            proc.wait()
            s_err.join()
            s_out.join()
            # Get buffers
            err = s_err.buffer
            out = s_out.buffer

        else:
            # Not piping needed...
            out, err = proc.communicate()
            out = "" if (out is None) else out.decode()
            err = "" if (err is None) else err.decode()

        # Retrieve execution result
        res = sp.CompletedProcess(args=proc.args, returncode=proc.returncode, stdout=out, stderr=err)

    # Store logs (if required)
    if log is not None:
        with log.open(mode="w", encoding="utf-8") as file:
            file.write(f"Date: {time.strftime('%Y/%m/%d - %H:%M:%S', time.localtime())}\n"
                       f"CWD : {cwd}\n"
                       f"CMD : {cmd}\n"
                       f"RET : {res.returncode}\n"
                       f"LOG : \n{out}\n{err}"
                       )

    # Return result
    return res



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Enumeration utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import enum


# -->> Definitions <<------------------


# -->> API <<--------------------------
class IntEnum(enum.IntEnum):
    """
    Integer enumeration class.
    Extends the functionalities of the base integer enumeration definition.
    """
    def __repr__(self) -> str:
        """
        Representation string.
        """
        return f"{self.__class__.__name__}({self.__str__()})"

    def __str__(self) -> str:
        """
        Class object as string.
        """
        return f"{self.name}(0x{self.value:X})"

    @classmethod
    def from_int(cls, value: int) -> "IntEnum":
        """
        Converts, if possible, the input value to an enum object.

        :param int value: Value to be converted.

        :returns: Enumeration object.
        :rtype: IntEnum

        :raises ValueError: Value is not contained on the enum.
        """
        if cls.has_value(value=value):
            return cls(value)
        raise ValueError(f"Value 0x{value:02X} is not defined.")

    @classmethod
    def has_value(cls, value: int) -> bool:
        """
        Checks if the input value exist on the enum definition.

        :param int value: Value to be checked.

        :returns: True if exists, false otherwise.
        :rtype: bool
        """
        # pylint: disable=E1101
        return value in getattr(cls, "_value2member_map_")



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Path checking utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import pathlib as pl
import typing as tp

from .common import TPAny, TPByte, TPPath


# -->> Definitions <<------------------


# -->> API <<--------------------------
class FileTypeError(OSError):
    """
    File type is not the expected.
    """


class Path(pl.Path):
    """
    Path class extensions.
    """
    def __new__(cls, *args, **kwargs) -> 'Path':
        """
        Extends the Path initialization to new supported types.

        :return: Path object.
        :rtype: pl.Path

        :raises TypeError:  Input type cant be converted to a path.
        """
        tp_byte = getattr(TPByte, "__constraints__")
        tp_path = getattr(TPPath, "__constraints__")

        # Avoid not compatible types
        path  = []
        for item in args:
            # Check type
            if not isinstance(item, tp_path):
                raise TypeError(f"Argument should be a compatible type ({tp_path}). {type(item)} is not supported.")
            # Convert
            if isinstance(item, tp_byte):
                path.append(bytes(item).decode(errors="ignore"))
            else:
                path.append(str(item))

        # Generate object and add extra functionalities
        obj = pl.Path(*tuple(path))
        setattr(obj.__class__, Path.reachable.__name__, Path.reachable)
        setattr(obj.__class__, Path.validate.__name__, staticmethod(Path.validate))
        setattr(obj.__class__, Path.validate_dir.__name__, staticmethod(Path.validate_dir))
        setattr(obj.__class__, Path.validate_file.__name__, staticmethod(Path.validate_file))

        return obj

    def reachable(self) -> bool:
        """
        Checks if the path is reachable.

        :returns: True if reachable, false otherwise.
        :rtype: bool
        """
        return self.exists() or self.parent.exists()

    @staticmethod
    def validate(
            path:       TPAny = None,
            none_ok:    bool = False,
            reachable:  bool = False,
            must_exist: bool = False
            ) -> tp.Optional['Path']:
        """
        Validates the provided input.

        :param TPAny path:          Path to be converted / validated.
        :param bool none_ok:        Allows None input.
        :param bool reachable:      Path must be reachable.
        :param bool must_exist:     Path must exist.

        :return: Verified path.
        :rtype: Path

        :raises TypeError:          Input type cant be converted to a path.
        :raises ValueError:         Provided path is not supported.
        :raises FileNotFoundError:  Path cant be reached or doesnt exist.
        """
        # Check if input is none
        if path is None:
            if none_ok:
                return None
            raise ValueError("Validation failed: None is not accepted as path.")

        # Validate
        path = Path(path)
        if reachable and not path.reachable():
            raise FileNotFoundError(f"Validation failed: {path} is not reachable.")
        if must_exist and not path.exists():
            raise FileNotFoundError(f"Validation failed: {path} doesnt exist.")

        return path

    @staticmethod
    def validate_dir(
            path:       TPAny = None,
            none_ok:    bool = False,
            must_exist: bool = False,
            create:     bool = False,
            ) -> tp.Optional['Path']:
        """
        Validate the directory path.

        :param TPAny path:          Path to be converted / validated.
        :param bool none_ok:        Allows None input.
        :param bool must_exist:     Path must exist.
        :param bool create:         create directory if doesn't exist.

        :return: Verified path.
        :rtype: Path

        :raises TypeError:          Input type cant be converted to a path.
        :raises ValueError:         Provided path is not supported.
        :raises PathError:          Path is not a directory.
        :raises FileNotFoundError:  Path cant be reached or doesnt exist.
        """
        # Validate base path
        path = Path.validate(path=path, none_ok=none_ok, reachable=True)
        if path is None:
            return None

        # Create
        if create and not path.exists():
            path.mkdir()

        # Validate
        if must_exist and not path.exists():
            raise FileNotFoundError(f"Validation failed: {path} doesnt exist.")
        if path.exists() and not path.is_dir():
            raise FileTypeError(f"Validation failed: {path} exists but is not a directory.")

        return path

    @staticmethod
    def validate_file(
            path:       TPAny = None,
            none_ok:    bool = False,
            must_exist: bool = False,
            default:    str = None
            ) -> tp.Optional['Path']:
        """
        Validate the file path.

        :param TPAny path:          Path to be converted / validated.
        :param bool none_ok:        Allows None input.
        :param bool must_exist:     Path must exist.
        :param str default:         Filename to be used when the input is a directory.

        :return: Verified path.
        :rtype: Path

        :raises TypeError:          Input type cant be converted to a path.
        :raises ValueError:         Provided path is not supported.
        :raises PathError:          Path is not a file.
        :raises FileNotFoundError:  Path cant be reached or doesnt exist.
        """
        # Validate base path
        path = Path.validate(path=path, none_ok=none_ok, reachable=True)
        if path is None:
            return None

        # Default
        if default and path.exists() and path.is_dir():
            path = path / default

        # Validate
        if must_exist and not path.exists():
            raise FileNotFoundError(f"Validation failed: {path} doesnt exist.")
        if path.exists() and not path.is_file():
            raise FileTypeError(f"Validation failed: {path} exists but is not a file.")

        return path



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Math utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import math


# -->> Definitions <<------------------


# -->> API <<--------------------------
def closest_multi(ref: float, base: float, force_next: bool = False) -> float:
    """
    Computes the closest multiple of base based on a given reference.

    :param float ref:       Reference.
    :param float base:      Multiple base.
    :param bool force_next: If reference is equal to multiple, forces next multiple.

    :return: Closest multiple.
    :rtype: float
    """
    mod = ref % base
    if mod != 0:
        return ref + (base - mod)
    return ref + base if force_next else ref


def closest_pow(ref: float, base: float, force_next: bool = False) -> float:
    """
    Computes the closest power of base based on a given reference.

    :param float ref:       Reference
    :param float base:      Exponential base
    :param bool force_next: If reference is equal to power of base, forces next power.

    :return: Closest power.
    :rtype: float
    """
    aux = round(math.log(ref, base))
    val = base ** aux
    if (val < ref) or (force_next and (val == ref)):
        val = base ** (aux + 1)
    return val



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Project build utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

from ..utils.path import TPPath, Path
from ..utils.subprocess import execute


# -->> Definitions <<------------------


# -->> API <<--------------------------
def build_cubeide(name: str, config: str, project: TPPath, workspace: TPPath, indexer: bool = False,
                  log: TPPath = None, pipe: bool = False
                  ) -> None:
    """
    Calls the STM32 CubeIDE headless builder on the specified project.

    :note:

        - stm32cubeidec executable must be in the PATH.
        - You cannot build if the workspace is already in use.
        - Reference: https://gnu-mcu-eclipse.github.io/advanced/headless-builds/

    :param str name:            Project name.
    :param str config:          Build configuration to be used. Ex: Debug, Release.
    :param TPPath project:      Path to CubeIDE project.
    :param TPPath workspace:    Path to CubeIDE workspace.
    :param bool indexer:        Runs the indexer on build.
    :param TPPath log:          Path to store the build logs.
    :param bool pipe:           Enable pipe output to terminal.
    """
    # Validate paths
    workspace   = Path.validate_dir(path=workspace)
    project     = Path.validate_dir(path=project, must_exist=True)
    log         = Path.validate_dir(path=log, none_ok=True)

    # Build
    idx = "-no-indexer" if not indexer else ""
    cmd = f"stm32cubeidec --launcher.suppressErrors -nosplash {idx} " \
          f"-application org.eclipse.cdt.managedbuilder.core.headlessbuild " \
          f"-data '{workspace}' -cleanBuild '{name}/{config}' -import '{project}'"
    res = execute(cmd=cmd, pipe=pipe, log=log)
    if not pipe and res.returncode:
        print(f"Command:\n{cmd}\nFailed with error:\n{res.stderr}")


def build_iar(config: str, project: TPPath,
              log: TPPath = None, pipe: bool = False
              ) -> None:
    """
    Calls the IAR headless builder on the specified project.

    :note:

        - IarBuild executable must be in the PATH.

    :param str config:          Build configuration to be used. Ex: Debug, Release.
    :param TPPath project:      Path to EWARM project.
    :param TPPath log:          Path to store the build logs.
    :param bool pipe:           Enable pipe output to terminal.
    """
    # Validate paths
    project = Path.validate_dir(path=project, must_exist=True)
    log = Path.validate_dir(path=log, none_ok=True)

    # Build
    cmd = f'IarBuild "{project}" -build "{config}"'
    res = execute(cmd=cmd, pipe=pipe, log=log)
    if not pipe and res.returncode:
        print(f"Command:\n{cmd}\nFailed with error:\n{res.stderr}")



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Serialized object abstract implementation.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import abc
import typing as tp


# -->> Definitions <<------------------


# -->> API <<--------------------------
class AbstractSerialized(abc.ABC):
    """
    Serialized object abstraction.
    This class implements the expected interface for a serialized object.
    """
    def __repr__(self) -> str:
        """
        Representation string.
        """
        return f"{self.__class__.__name__}(serialized=0x{self.serialize().hex()})"

    def __eq__(self, other: object) -> bool:
        """
        Check if the object is equal to the input.
        """
        if isinstance(other, self.__class__):
            return self.serialize() == other.serialize()
        return False

    def __ne__(self, other: object):
        """
        Check if the object is different to the input.
        """
        return not self.__eq__(other)

    @abc.abstractmethod
    def serialize(self) -> bytearray:
        """
        Serializes the item into a bytearray.

        :returns: Serialized object.
        :rtype: bytearray.
        """

    @classmethod
    @abc.abstractmethod
    def deserialize(cls, data: bytearray) -> tp.Optional["AbstractSerialized"]:
        """
        Deserializes an object from a bytearray.

        :param bytearray data: Data to extract the object from.

        :returns: Deserialized object if available, None otherwise.
        :rtype: Optional["AbstractSerialized"]
        """


class AbstractSerializedCodec(abc.ABC):
    """
    Serialized object codec abstraction.
    This class implements the logic used to encode/decode a serialized object.
    """
    @abc.abstractmethod
    def encode(self, data: AbstractSerialized) -> bytearray:
        """
        Encodes a serialized object into a byte array.

        :param AbstractSerialized data: Object to encode.

        :returns: Encoded serialized data.
        :rtype: bytearray
        """

    @abc.abstractmethod
    def decode(self, data: bytearray) -> tp.Optional[AbstractSerialized]:
        """
        Decodes a serialized object from a byte array.

        :param bytearray data: Bytes to decode.

        :returns: Deserialized object if able, None otherwise.
        :rtype: Optional[AbstractSerialized]
        """



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Binary generation utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import typing as tp

import intelhex

from .path import TPPath, Path


# -->> Definitions <<------------------
RECORD_BYTES = 32


# -->> API <<--------------------------
def bin_to_hex(src: TPPath, out: TPPath = None, off: int = 0x08000000) -> intelhex.IntelHex:
    """
    Convert a binary file into an HEX.

    :param TPPath src: Path to source BIN file.
    :param TPPath out: Optional. If provided, saves the generated hex file in the given path.
    :param int off:    BIN file address offset.

    :return: HEX file
    :rtype: intelhex.IntelHex
    """
    # Check paths
    src = Path.validate_file(path=src, none_ok=False, must_exist=True)
    out = Path.validate_file(path=out, none_ok=True)
    # Generate HEX
    tmp = intelhex.IntelHex()
    tmp.loadbin(fobj=f"{src}", offset=off)
    # Save if required
    if out is not None:
        tmp.write_hex_file(f=f"{out}", byte_count=RECORD_BYTES)
    return tmp


def merge_bin(src: tp.List[tp.Tuple[TPPath, int]], out: TPPath = None) -> intelhex.IntelHex:
    """
    Merge a group of binary files into an HEX.

    :param list src:    List of tuples that contain:
        - BIN file path.
        - BIN file address offset
    :param TPPath out:  Optional. If provided, saves the generated hex file in the given path.

    :return: Merged HEX file.
    :rtype: intelhex.IntelHex
    """
    # Check paths
    src = [[Path.validate_file(path=path, none_ok=False, must_exist=True), addr] for path, addr in src]
    out = Path.validate_file(path=out, none_ok=True)
    # Merge all BIN files
    tmp = intelhex.IntelHex()
    for file, addr in src:
        this = bin_to_hex(src=file, off=addr)
        tmp.merge(other=this, overlap="replace")
    # Save if required
    if out is not None:
        tmp.write_hex_file(f=f"{out}", byte_count=RECORD_BYTES)
    return tmp


def merge_hex(src: tp.List[TPPath], out: TPPath = None) -> intelhex.IntelHex:
    """
    Merge a group of HEX files.

    :param list src:    List of paths to HEX files.
    :param TPPath out:  Optional. If provided, saves the generated hex file in the given path.

    :return: Merged HEX file.
    :rtype: intelhex.IntelHex
    """
    # Check paths
    src = [Path.validate_file(path=path, none_ok=False, must_exist=True) for path in src]
    out = Path.validate_file(path=out, none_ok=True)
    # Merge all HEX files
    tmp = intelhex.IntelHex()
    for file in src:
        this = intelhex.IntelHex(source=f"{file}")
        tmp.merge(other=this, overlap="replace")
    # Save if required
    if out is not None:
        tmp.write_hex_file(f=f"{out}", byte_count=RECORD_BYTES)
    return tmp



from .build import build_cubeide, build_iar
from .version import (
    VersionHandler,
    AbstractVersionUpdater, AbstractVersionExporter, AbstractVersionStorage,
    GitBuildVersionUpdater, CCppVersionExporter, SimpleVersionStorage,
    )



#!/usr/bin/python
# -*- coding: ascii -*-
"""
<TBD>

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import os
import sys

from .common import TPAny
from .path import Path

# -->> Tunables <<---------------------


# -->> Definitions <<------------------


# -->> API <<--------------------------
def activate(venv: TPAny) -> None:
    """
    Activates the given virtual environment on the current run.

    :param TPAny venv:          Path to virtualenv folder.

    :raises TypeError:          Input type cant be converted to a path.
    :raises ValueError:         Provided path is not supported.
    :raises PathError:          Path is not a file.
    :raises FileNotFoundError:  Path cant be reached or doesnt exist.
    :raises EnvironmentError:   Virtualenv not active or it doesn't match target.
    """
    # Path Validation
    venv    = Path(venv)
    script  = Path.validate_file(path=venv / "Scripts/activate_this.py", must_exist=True)

    # Activation
    with script.open(mode="r", encoding="utf-8") as file:
        code = compile(file.read(), script, "exec")
        exec(code, dict(__file__=script))

    # Validation
    active = Path(os.environ["VIRTUAL_ENV"]) if ("VIRTUAL_ENV" in os.environ) else sys.executable
    if venv != active:
        raise EnvironmentError(f"Active environment doesn't match target: {venv} != {active}.")



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Timing utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import functools as fc
import time
import typing as tp

from .common import TPAny, CBAny2Any


# -->> Definitions <<------------------


# -->> API <<--------------------------
def timer(name: str = None, logger: CBAny2Any = print, precision: int = 6) -> tp.Callable[[CBAny2Any], CBAny2Any]:
    """
    Decorator.
    Used to log the execution time of a function.

    :param str name:            Timer name. Used to identify the measurement.
    :param CBAny2Any logger:    Callable used to print the time logs. By default, system print.
    :param float precision:     Number of decimals included on elapsed time.

    :return: Decorated function wrapped on timer.
    :rtype: Callable[[Callable[..., RT]], Callable[..., RT]]
    """
    def decorator(func: CBAny2Any) -> CBAny2Any:
        @fc.wraps(func)
        def wrapper(*args, **kwargs) -> TPAny:
            tim = Timer()
            ret = func(*args, **kwargs)
            tot = tim.elapsed()
            if logger is not None:
                logger(f"{name}: Execution time: {tot:.{precision}f}[s]")
            return ret
        return wrapper
    return decorator


class Timer:
    """
    Simple process timer.
    """
    def __init__(self) -> None:
        """
        This class don't require any input from the user to be initialized.
        """
        self._start = 0.0
        self.start()

    def start(self) -> float:
        """
        Updates the timer start timestamp.

        :returns: Start time.
        :rtype: float
        """
        self._start = time.perf_counter()
        return self._start

    def elapsed(self, update: bool = False) -> float:
        """
        Computes the time elapsed since the latest timer start.

        :param bool update: Computes elapsed time and updated the start time.

        :returns: Time elapsed.
        :rtype: float
        """
        now = time.perf_counter()
        ret = now - self._start
        if update:
            self._start = now
        return ret



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Event handling utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

from .common import CBAny2None


# -->> Definitions <<------------------


# -->> API <<--------------------------
class EventHook:
    """
    Simple event hook implementation.
    This utility allows to emit events to several callbacks in a simple fashion.

    .. note::
        *   All the callbacks subscribed to a hook need to handle the same arguments.
    """

    def __init__(self) -> None:
        """
        This class don't require any input from the user to be initialized.
        """
        self._callbacks = []

    def __iadd__(self, callback: CBAny2None) -> "EventHook":
        """
        Simplified callback subscription. Overrides the **+=** operator.

        :param callable callback: Event callback.

        :returns: Self.
        :rtype: EventHook
        """
        self.subscribe(callback=callback)
        return self

    def __isub__(self, callback: CBAny2None) -> "EventHook":
        """
        Simplified callback unsubscription. Overrides the **-=** operator.

        :param callable callback: Event callback.

        :returns: Self.
        :rtype: EventHook
        """
        self.unsubscribe(callback=callback)
        return self

    @property
    def empty(self) -> bool:
        """
        Returns if the event hook has callbacks subscribed.
        """
        return len(self._callbacks) == 0

    def subscribe(self, callback: CBAny2None) -> bool:
        """
        Subscribes a callback to the event hook.

        :param callable callback: Event callback.

        :returns: True if subscribed, false otherwise.
        :rtype: bool
        """
        if callback not in self._callbacks:
            self._callbacks.append(callback)
            return True
        return False

    def unsubscribe(self, callback: CBAny2None) -> bool:
        """
        Unsubscribes a callback from the event hook.

        :param callable callback: Event callback.

        :returns: True if unsubscribed, false otherwise.
        :rtype: bool
        """
        if callback in self._callbacks:
            self._callbacks.remove(callback)
            return True
        return False

    def clear(self) -> None:
        """
        Remove all the callbacks from the hook.
        """
        self._callbacks.clear()

    def emit(self, *args, **kwargs) -> None:
        """
        Emit the given arguments to all the callbacks.

        :param args: Arguments.
        :param kwargs: Key arguments.
        """
        for callback in self._callbacks:
            callback(*args, **kwargs)



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Common SDK definitions.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import pathlib as pl
import typing as tp


# -->> Tunables <<---------------------


# -->> Definitions <<------------------
#: TyPe definition. Any value.
TPAny       = tp.TypeVar("TPAny")

#: TyPe definition. Path sources.
TPByte      = tp.TypeVar("TPByte", bytes, bytearray)

#: TyPe definition. Path sources.
TPText      = tp.TypeVar("TPText", bytes, bytearray, str)

#: TyPe definition. Path sources.
TPPath      = tp.TypeVar("TPPath", bytes, bytearray, str, pl.Path)

#: CallBack definition. Any -> Any
CBAny2Any   = tp.Callable[..., TPAny]

#: CallBack definition. Any -> None
CBAny2None  = tp.Callable[..., None]

#: CallBack definition. None -> None
CBNone2None = tp.Callable[[], None]


# -->> API <<--------------------------



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Serial interface implementation. This class represents the closer point to the
user. Here the developer needs to complete the interface with the specific
applications commands.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import time
import typing as tp

from ..utils.logger import SDK_LOG
from ..utils.serialized import AbstractSerialized
from ..utils.time import Timer
from .stream import Stream


# -->> Definitions <<------------------
#: CallBack definition. AbstractSerialized -> bool
CBSerialized2Bool = tp.Callable[[AbstractSerialized], bool]


# -->> API <<--------------------------
class Interface:
    """
    Serial command interface implementation.
    This class should implement all the methods to interact with the target device.

    Available events:

    #.  **on_receive:** This event is emitted when an object is received and
        deserialized from the serial device. Subscribe using callbacks with
        syntax::

            def <callback>(item: AbstractSerialized)

    #.  **on_connect:** This event is emitted when the system is able to
        connect to the device. Subscribe using callbacks with syntax::

            def <callback>()

    #.  **on_reconnect:** This event is emitted when the system is able to
        reconnect to the device. Subscribe using callbacks with syntax::

            def <callback>()

    #.  **on_disconnect:** This event is emitted when the system gets
        disconnected from the device. Subscribe using callbacks with syntax::

            def <callback>()

    """
    #: Interface command pull period
    PERIOD_PULL_S = 0.005
    #: Interface command response timeout
    TIMEOUT_RESPONSE_S = 0.5

    def __init__(self, stream: Stream) -> None:
        """
        Class initialization.

        :param Stream stream:   Stream used to run the interface.
        """
        # Response timeout configuration
        self._timeout = self.TIMEOUT_RESPONSE_S
        # Initialize stream
        self._stream = stream
        # Attach stream events
        self.on_connect     = self._stream.on_connect
        self.on_reconnect   = self._stream.on_reconnect
        self.on_disconnect  = self._stream.on_disconnect
        self.on_receive     = self._stream.on_receive
        SDK_LOG.info(f"Interface initialized on: {self._stream.device}")

    @property
    def stream(self) -> Stream:
        """
        Stream handler.
        """
        return self._stream

    @property
    def timeout(self) -> float:
        """
        Message response timeout.
        """
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: float) -> None:
        """
        Message response timeout setter.

        :param float timeout: Timeout in seconds.
        """
        if timeout <= 0.0:
            raise ValueError("The response timeout needs to be greater than zero.")
        self._timeout = timeout

    def transmit(self,
                 send: AbstractSerialized,
                 logic: tp.Optional[CBSerialized2Bool] = None,
                 timeout: float = None) -> tp.Optional[AbstractSerialized]:
        """
        Send a item and, if required, wait for a response that complies with
        the defined logic.

        The response detection logic should have the following syntax::

            def <callback>(AbstractSerialized) -> bool:

        Where the input is the item received from the device.

        :param AbstractSerialized send:     Packet to be sent.
        :param CBSerialized2Bool logic:     Response logic. Defines if a response is detected. If none then only sends.
        :param float timeout:               Response timeout. By default the interface setting.

        :returns: None if timeout or no response is detected, response item otherwise.
        :rtype: Optional[AbstractSerialized]
        """
        if logic:
            return self._send_recv(send=send, logic=logic, timeout=timeout)
        self._send_none(send=send)
        return None

    def stop(self) -> None:
        """
        Stops the stream process.
        """
        self._stream.stop()

    def join(self) -> None:
        """
        Maintains the process alive while the interface is working. The loop will
        exit if the interface serial process gets stopped.
        """
        self._stream.join()

    def _send_none(self,
                   send: AbstractSerialized) -> None:
        """
        Sends an item and dont care about the response.

        :param AbstractSerialized send:     Packet to be sent.
        """
        # Send item and return
        SDK_LOG.debug("Sending item...")
        self._stream.send(item=send)
        SDK_LOG.debug("Response not needed.")

    def _send_recv(self,
                   send: AbstractSerialized,
                   logic: CBSerialized2Bool,
                   timeout: float = None) -> tp.Optional[AbstractSerialized]:
        """
        Sends an item and wait for response.

        :param AbstractSerialized send:     Item to be sent.
        :param CBSerialized2Bool logic:     Response logic. Defines if a response is detected. If none then only sends.
        :param float timeout:               Response timeout. By default the interface setting.

        :returns: None if timeout or no response is detected, response item otherwise.
        :rtype: Optional[AbstractSerialized]
        """
        # Prepare data
        recv = None
        timeout = self._timeout if (timeout is None) else timeout

        # Receive logic callback
        def on_received(item: AbstractSerialized) -> None:
            nonlocal recv
            if logic(item):
                recv = item

        # Prepare response logic
        self.on_receive += on_received

        # Prepare and send
        SDK_LOG.debug("Sending item...")
        self._stream.send(item=send)

        # Wait for response
        SDK_LOG.debug("Waiting for response...")
        tim = Timer()
        while not recv and (tim.elapsed() < timeout):
            time.sleep(self.PERIOD_PULL_S)
        self.on_receive -= on_received

        # Check data
        state = "Received" if recv else "Timeout"
        SDK_LOG.debug(f"Item response: {state} after {tim.elapsed():.03f}[s]")
        return recv



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Byte manipulation utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""


# -->> Definitions <<------------------


# -->> API <<--------------------------
def bitmask(bit: int, fill: bool = False) -> int:
    """
    Returns a bitmask for the given bit length.

    :param int bit:     Bit length.
    :param bool fill:   If true, fill the mask with ones.

    :returns: bitmask.
    :rtype: int
    """
    mask = (1 << bit)
    return (((mask - 1) << 1) | 1) if fill else mask


def reverse_bits(value: int, size: int = None) -> int:
    """
    Reverse the bits of the input value.

    :param int value:   Value to be reversed.
    :param int size:    Size of the input in bits. By default it computes the minimum number of bits.

    :returns: Reversed value.
    :rtype: int
    """
    if not size:
        size = value.bit_length()
    val = value & bitmask(bit=(size - 1), fill=True)
    aux = f"{val:0{size:d}b}"
    return int(aux[::-1], base=2)


def reverse_bytes(value: int, size: int = None) -> int:
    """
    Reverse the bytes of the input value.

    :param int value:   Value to be reversed.
    :param int size:    Size of the input in bytes. By default it computes the minimum number of bytes.

    :returns: Reversed value.
    :rtype: int
    """
    if not size:
        size = (value.bit_length() + 7) // 8
    val = value & bitmask(bit=((8 * size) - 1), fill=True)
    aux = val.to_bytes(length=size, byteorder="big")
    return int.from_bytes(bytes=aux[::-1], byteorder="big")



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Service implementation.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import abc
import threading as th
import time

from .logger import SDK_LOG
from .threading import SDK_TP, SimpleThreadTask


# -->> Definitions <<------------------


# -->> API <<--------------------------
class AbstractService(abc.ABC):
    """
    Service definition abstraction.
    Use this class to simplify the service definition and handling.
    """
    #: Task execution delay.
    TASK_DELAY_S = 0.001

    def __init__(self, delay: float = TASK_DELAY_S) -> None:
        """
        Class initialization.

        :param float delay:     Time between service task executions.
        """
        # Configure service operation
        self._delay     = delay
        self._alive     = True
        self._resumed   = False
        self._ended     = th.Event()        # Set when the service gets terminated.
        self._running   = th.Event()        # Set when the service is running.
        self._executed  = th.Event()        # Set after every main task execution.
        # Start
        self._running.set()
        SDK_TP.enqueue(task=SimpleThreadTask(
            name=f"{self.__class__.__name__}",
            task=self._service
            ))

    def __del__(self):
        """
        Class destructor.
        Ensures to stop the service correctly on deletion.
        """
        self.stop()
        self.join()

    @property
    def is_alive(self) -> bool:
        """
        Returns if the service is alive.
        """
        return self._alive

    @property
    def is_running(self) -> bool:
        """
        Returns if the service is running.
        """
        return self._alive and self._running.is_set()

    @property
    def delay(self) -> float:
        """
        Service main task execution delay in seconds.
        """
        return self._delay

    @delay.setter
    def delay(self, value: float) -> None:
        """
        Service main task execution delay setter.

        :param float value: Delay in seconds.
        """
        self._delay = value

    def resume(self) -> None:
        """
        Resumes the service execution.
        """
        if not self._running.is_set():
            self._resumed = True
            self._running.set()
            SDK_LOG.info(f"{self.__class__.__name__} resumed")

    def pause(self) -> None:
        """
        Pauses the service execution.
        """
        if self._running.is_set():
            self._executed.wait()
            self._running.clear()
            self._on_pause()
            SDK_LOG.info(f"{self.__class__.__name__} paused")

    def stop(self) -> None:
        """
        Stops the service execution.
        """
        self._alive   = False
        self._resumed = False
        self._running.set()

    def join(self) -> None:
        """
        Wait until service is fully terminated.
        """
        self._ended.wait()

    def _service(self) -> None:
        """
        Service process execution.

        #. Runs the starting code.
        #. Run the execution loop until the service stop is requested.
        #. Runs the ending code and return.
        """
        # Start service execution
        SDK_LOG.info(f"{self.__class__.__name__} started")
        self._on_start()
        # Execution loop
        while self._alive:
            # Handle pause / resume
            self._running.wait()
            if self._resumed:
                self._on_resume()
                self._resumed = False
            # Handle termination
            if not self._alive:
                break
            # Run main task
            try:
                self._executed.clear()
                self._task()
            finally:
                self._executed.set()
                time.sleep(self._delay)
        # Run service ending code
        self._on_end()
        self._ended.set()
        SDK_LOG.info(f"{self.__class__.__name__} ended")

    @abc.abstractmethod
    def _task(self) -> None:
        """
        Service core task/logic.
        """

    def _on_start(self) -> None:
        """
        Optional. Called on service start.
        """

    def _on_resume(self) -> None:
        """
        Optional. Called on service resume.
        """

    def _on_pause(self) -> None:
        """
        Optional. Called on service pause.
        """

    def _on_end(self) -> None:
        """
        Optional. Called upon service termination.
        """



from .binary import bin_to_hex, merge_bin, merge_hex
from .bytes import bitmask, reverse_bits, reverse_bytes
from .cobs import COBS
from .common import TPAny, TPByte, TPPath, TPText, CBAny2Any, CBAny2None, CBNone2None
from .crc import CRC
from .enum import IntEnum
from .events import EventHook
from .logger import SDK_LOG, Logger
from .math import closest_multi, closest_pow
from .path import Path, FileTypeError
from .serialized import AbstractSerialized, AbstractSerializedCodec
from .service import AbstractService
from .stream import StreamRedirect
from .subprocess import execute
from .threading import SDK_TP, AbstractThreadTask, SimpleThreadTask, ThreadPool, get_threads, sync
from .time import Timer, timer
from .venv import activate
from .version import Version



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Stream implementation.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import abc
import time
import typing as tp

from ..utils.events import EventHook
from ..utils.logger import SDK_LOG
from ..utils.threading import SDK_TP, SimpleThreadTask
from ..utils.serialized import AbstractSerialized, AbstractSerializedCodec
from ..utils.service import AbstractService
from .device import Device


# -->> Definitions <<------------------


# -->> API <<--------------------------
class AbstractSerializedStreamCodec(AbstractSerializedCodec):
    """
    AbstractSerializedCodec variant for stream usage.
    This class includes the logic required to decode a serialized object from a byte stream.
    """
    @abc.abstractmethod
    def decode_stream(self, device: Device) -> tp.Optional[AbstractSerialized]:
        """
        Decodes a serialized object from a byte stream.

        :param Device device: Stream source.

        :returns: Deserialized object if able, None otherwise.
        :rtype: Optional[AbstractSerialized]

        :raises ConnectionError: Device raised exception while reading.
        """


class Stream(AbstractService):
    """
    This class is used to send and receive serialized objects through a serial device
    in asynchronous way. When a new item is received this class will notify the system
    using events.

    Available events:

    #.  **on_receive:** This event is emitted when an object is received and
        deserialized from the serial device. Subscribe using callbacks with
        syntax::

            def <callback>(item: AbstractSerialized)

    #.  **on_connect:** This event is emitted when the system is able to
        connect to the device. Subscribe using callbacks with syntax::

            def <callback>()

    #.  **on_reconnect:** This event is emitted when the system is able to
        reconnect to the device. Subscribe using callbacks with syntax::

            def <callback>()

    #.  **on_disconnect:** This event is emitted when the system gets
        disconnected from the device. Subscribe using callbacks with syntax::

            def <callback>()

    """
    #: Stream reconnect period
    RECONNECT_PERIOD_S = 0.5

    def __init__(self, device: Device, codec: AbstractSerializedStreamCodec, **kwargs) -> None:
        """
        Class initialization.

        :param Device device:                       Serial device handler.
        :param AbstractSerializedStreamCodec codec: Serialized objects codec handler.
        """
        # Service core
        super().__init__(**kwargs)
        self._device        = device
        self._codec         = codec
        # Service events
        self.on_receive     = EventHook()
        self.on_connect     = EventHook()
        self.on_reconnect   = EventHook()
        self.on_disconnect  = EventHook()
        # Internal event handlers
        self.on_receive     += lambda item: self._transfer_debug(item=item, received=True)

    def __del__(self) -> None:
        """
        Class destructor.
        Stops the stream and the associated serial device.
        """
        super().__del__()
        del self._device

    @property
    def codec(self) -> AbstractSerializedStreamCodec:
        """
        Serialized objects codec handler.
        """
        return self._codec

    @property
    def device(self) -> Device:
        """
        Serial device handler.
        """
        return self._device

    @device.setter
    def device(self, device: Device) -> None:
        """
        Serial device handler setter.

        .. note::
            *   This action replaces the current device.
            *   If the new device handler has no port defined
                the current port is preserved.

        :param Device device: Serial device handler.
        """
        self.pause()
        if device.port is not None:
            self._device = device
        self.resume()

    def send(self, item: AbstractSerialized) -> None:
        """
        Send a serializable item through the serial device.

        :param AbstractSerialized item: Item to send.
        """
        if self.is_running:
            self._transfer_debug(item=item, received=False)
            self._device.write(data=self._codec.encode(data=item))

    def _task(self) -> None:
        """
        Stream process:

        #. Tries to decode items.
        #. If an item is available, emits an event.
        #. Handle disconnection status.

        """
        # Try to receive data, if failed go to reconnection loop
        try:
            item = self._codec.decode_stream(device=self._device)
            if item:
                SDK_TP.enqueue(task=SimpleThreadTask(
                    name=f"{self.__class__.__name__}.on_received",
                    task=lambda: self.on_receive.emit(item=item)
                    ))

        except ConnectionError:
            # Device disconnected...
            SDK_LOG.info(f"Device disconnected: {self._device}")
            SDK_TP.enqueue(task=SimpleThreadTask(
                name=f"{self.__class__.__name__}.on_disconnect",
                task=self.on_disconnect.emit
                ))
            if self._device_connect():
                SDK_TP.enqueue(task=SimpleThreadTask(
                    name=f"{self.__class__.__name__}.on_reconnect",
                    task=self.on_reconnect.emit
                    ))

    def _on_start(self) -> None:
        """
        Ensures device connected on service start.
        """
        self._device_init()

    def _on_resume(self) -> None:
        """
        Ensures device connected on service resume.
        """
        self._device_init(start=False)

    def _on_pause(self) -> None:
        """
        Closes device when paused.
        """
        self._device.close()

    def _on_end(self) -> None:
        """
        Closes device when service gets terminated.
        """
        self._device.close()

    def _device_init(self, start: bool = True) -> None:
        """
        Initializes the serial device.

        :param bool start:  True if used on start handler. False otherwise.
        """
        if not self._device.is_open:
            connected = self._device_connect()
            if connected and start:
                SDK_TP.enqueue(task=SimpleThreadTask(
                    name=f"{self.__class__.__name__}.on_connect",
                    task=self.on_connect.emit
                    ))

    def _device_connect(self) -> bool:
        """
        Perform connection attempts on the serial device while the service is running.

        :returns: True if connection succeeded, false otherwise.
        :rtype: bool
        """
        SDK_LOG.info(f"Attempting connection to {self._device}")
        self._device.close()
        while self.is_running:
            if self._device.open():
                self._device.flush()
                SDK_LOG.info(f"Connected to {self._device}")
                return True
            SDK_LOG.info(f"Connection attempt on {self._device} failed.")
            time.sleep(self.RECONNECT_PERIOD_S)
        return False

    @staticmethod
    def _transfer_debug(item: AbstractSerialized, received: bool) -> None:
        """
        Print the sent/received items on the logger:

        :param AbstractSerialized item: Item that is being sent/received.
        :param bool received:           Flag to indicate if we are sending/receiving the item.
        """
        action = "recv" if received else "sent"
        SDK_LOG.debug(f"Item {action}: {item}")
